#include <vector>
#include <set>
#include <map>
#include <queue>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <sys/stat.h>
#include <algorithm>
#include "gfakluge/gfakluge.hpp"
#include <time.h>

using namespace std;
using namespace gfak;

typedef vector<vector<size_t> > adj_list;
typedef vector<pair<size_t, size_t> > edge_list;
typedef vector<size_t> level_vector;
typedef vector<size_t> path_ends;
typedef vector<set<size_t>> cand_set;

typedef vector<map<size_t,size_t> > edge_flow;
typedef vector<size_t> node_flow;


// Needed Constants and Variables
const bool in = true;
const bool out = false;
const size_t NOT_SET = (size_t) -2;
const size_t NOT_FOUND = (size_t) -1;
size_t IDX_ONE = 0;
bool USE_COORD = false;

const unsigned int INIT_LEVEL = 0;
const unsigned int INIT_LEVEL_PLUS_SH = 1;
const unsigned int INIT_GSC = 2;

stringstream PATHCOVER_STREAM;
stringstream COORDS_STREAM;
stringstream DOT_STREAM;
stringstream LOG_STREAM;
vector<char> NODE_NAME;
vector<string> NODE_IDX;


// Function Definitions
tuple<level_vector,edge_flow,node_flow> calc_init_path_cover_level	(adj_list &adj, adj_list &adj_prev, size_t s, size_t t, vector<size_t> &indegree);
tuple<level_vector,edge_flow,node_flow> calc_init_path_cover_level_sh	(adj_list &adj, adj_list &adj_prev, size_t s, size_t t, vector<size_t> &indegree);
tuple<level_vector,edge_flow,node_flow> calc_init_path_cover_gsc		(adj_list &adj, adj_list &adj_prev, size_t s, size_t t, vector<size_t> &indegree);
void fill_flow_s_v												(size_t s,size_t v, adj_list &adj_prev, edge_flow &e_f, node_flow &v_f);
void fill_flow_u_v												(size_t u,size_t v, adj_list &adj_prev, level_vector &level ,edge_flow &e_f, node_flow &v_f, size_t p, cand_set &R);
tuple<edge_flow,node_flow> minimize_flow						(adj_list &adj, adj_list &adj_prev, size_t s, size_t t, edge_flow &adj_f, node_flow &v_f, size_t capacity);
vector<size_t>			 get_paths							(adj_list &adj, 				 size_t s, size_t t, size_t width, edge_flow &adj_f);



/**
 * Calculate the minimum pathcover and the level coordinate for each node
 * for the DAG G = (V, E), by running the following algorithms:
 *   1. calc_init_path_cover
 *   2. minimize_flow
 *   3. get_paths
 *
 * @param n  number of nodes in DAG G
 * @param edges     DAG G in form of an edge list
 * @param s         the source node of G
 * @param t         the sink node of G
 * @param init_method	indicator with method to initialize the flow network
 * @param silent    if true, no output will be printed
**/
void calculate_coordinate_system(size_t n, edge_list &edges, size_t s, size_t t, unsigned int init_method, bool silent) {
	
	//calculet adjacency list from edge list
	adj_list adj(n), adj_prev(n);
	vector<size_t> indegree(n);
	for (pair<size_t, size_t> e : edges) {
		size_t u = e.first;
		size_t v = e.second;
		adj[u].push_back(v);
		adj_prev[v].push_back(u);
		indegree[v]++;
	}
    
	time_t time1 = time(NULL);
	time_t time2;
    
	level_vector level;

    
	edge_flow adj_f, adj_prev_f;
	node_flow v_f;
	size_t capacity;
    
	
	//calculate initial path cover fill flow network
	clock_t init_time0 = clock();
	if(!silent) cout << "calc initial path cover ";
	switch(init_method){
		case INIT_LEVEL: 
			if(!silent) cout << "with level wise covering ..." << flush;
			tie(level, adj_f, v_f) = calc_init_path_cover_level(adj, adj_prev,s, t , indegree);
			break;
		case INIT_LEVEL_PLUS_SH: 
			if(!silent) cout << "with level wise covering + simple heuristic ..." << flush;
			tie(level, adj_f, v_f) = calc_init_path_cover_level_sh(adj, adj_prev,s, t , indegree);
			break;
		case INIT_GSC: 		
			if(!silent) cout << "with greedy set cover..." << flush;		
			tie(level, adj_f, v_f) = calc_init_path_cover_gsc(adj, adj_prev,s, t , indegree);
			break;
	}	
	time2 = time(NULL);
	if(!silent) cout << "\rTime for calculating initial path cover: " << difftime(time2,time1) << " s                              " << endl;
	if(!silent) cout << "Size of initial path cover: " << v_f[s] << endl;	
	time1 = time2;
	capacity = v_f[s];	
	clock_t init_time1 = clock();   
	LOG_STREAM << capacity << '\t'; //size of initial path cover
	
	//minimize flow 
	if(!silent) cout << "minimze path cover..." << flush;
	tie(adj_f, v_f) = minimize_flow(adj, adj_prev, s, t, adj_f, v_f, capacity); //adj_f altered
	time2 = time(NULL);
	if(!silent) cout << "\rTime for minimizing flow: " << difftime(time2,time1) << " s" << endl;
	time1 = time2;

	//get paths from flow network
	if(!silent) cout << "read paths from network..." << flush;
	size_t width = v_f[s];
	vector<size_t> pid;
	pid = get_paths(adj, s, t, width ,adj_f); //adj_f altered
	time2 = time(NULL);
	if(!silent) cout << "\rTime for reading paths of flow network: " << difftime(time2,time1) << " s" << endl;
	time1 = time2;
	if(!silent) cout << "Size of minimal path cover: " << width << endl;
	LOG_STREAM << width << '\t' <<  double(init_time1 - init_time0) / CLOCKS_PER_SEC << '\t';
		

	/*** OUTPUT ***/
	size_t vertices = adj.size();
	// print maxLevel coordinates
	COORDS_STREAM << "id, original id, coordinate" << endl;
	for (size_t i = 0; i < vertices; i++)
		COORDS_STREAM << i+IDX_ONE << ", " << NODE_IDX[i] <<", (" << level[i]+IDX_ONE << "," << pid[i]+IDX_ONE << ")" << endl;

	// print dot-graph
	DOT_STREAM << "digraph nucleotide_graph {\n\tnode [shape=circle];\n\trankdir=LR;\n\n";
	for (size_t i = 0; i < vertices; i++) {  // print nodes
		DOT_STREAM << "\t" << i+IDX_ONE << " [label=\""
				//<< "id=" << i+IDX_ONE << "\\n"
				//<< "seqnr=" << NODE_IDX[i] << "\\n"
				//<< i+IDX_ONE << "\\n"
				<< NODE_NAME[i] //<< "\\n"
				//<< "(" << level[i]+IDX_ONE << "," << pid[i]+IDX_ONE << ")"
				<< "\"";
		if (USE_COORD)
			DOT_STREAM << ", pos=\"" <<  level[i]+IDX_ONE << "," <<  ((int) (pid[i]+IDX_ONE)) << "!\"";
		
		DOT_STREAM << "];\n";
	}
	DOT_STREAM << "\n";
	for (size_t i = 0; i < vertices; i++) {  // print edges
		for (size_t w : adj[i]) {
			DOT_STREAM << "\t" << i+IDX_ONE << " -> " << w+IDX_ONE
					// << " [label=\"" << adj_f[i][w] << "\"]"
					<< ";\n";
		}
	}
	DOT_STREAM << "}" << "\n";
	return;
}




/**
 * Calculate the level for each node and a path cover that will be represented n a flow network.
 * 
 * initial path cover found with Algorithm 1 of the paper
 *
 * @param adj       DAG G in form of an adjacency list
 * @param adj_prev  the corresponding inverse adjacency list
 * @param s         the source node of G
 * @param t         the sink node of G
 * @param indegree  vector of size |V| containing the input degree for each node
 *
 * @return          tuple<level_vector, edge_flow, node_flow> with
 *                  level_vector: a vector containing the level of each node
 * 				edge_flow, node_flow: a flow network that represents the path cover
 */
tuple<level_vector,edge_flow,node_flow> calc_init_path_cover_level(adj_list &adj, adj_list &adj_prev, size_t s,size_t t, vector<size_t> & indegree) {
	size_t n = adj.size();
	size_t l = 0;
	size_t p = 0;
	deque<size_t> Ql;
	Ql.push_back(s);
	
	level_vector level(n);
	level[s] = 0;
	
	cand_set R(n);
	path_ends path(0);
    
	edge_flow e_f(n);
	node_flow v_f(n);
	
	unsigned int maxsize = 0;

	while (!Ql.empty()) {	
		//prepare Q for next level
		deque<size_t> Ql1(0);	   
		
		for (size_t v : Ql) {  		
			// calculate R_v
			for (size_t u : adj_prev[v]) {
				for(size_t i : R[u]){
					size_t last = path[i];
					if(level[last] < level[u] || last == u )
						R[v].emplace(i);
				}
			}
				
			if (R[v].empty()) {  
				// add new path id
				R[v].insert(p++);
				path.push_back(v);
				fill_flow_s_v(s,v,adj_prev, e_f, v_f); 
			}else{
				size_t i = *(R[v].begin());
				// find path from last to v
				size_t last = path[i];			
				fill_flow_u_v(last,v,adj_prev, level ,e_f, v_f, i, R);
				path[i] = v;
			}

			for (size_t w : adj[v]) {
				indegree[w]--;
				if (!indegree[w]) {
					level[w] = l + 1;
					Ql1.push_back(w);
				}
			}
		}
		l++;
		swap(Ql,Ql1);
	}
	
	for(unsigned int i = 0; i < path.size() ; i++)
		fill_flow_s_v(path[i],t,adj_prev, level ,e_f, v_f, i, R);   
	
	return make_tuple(level,e_f,v_f);
}

/**
 * Calculate the level for each node and a path cover that will be represented n a flow network.
 * 
 * initial path cover found with Algorithm 1 of the paper + simple heuristic
 *
 * @param adj       DAG G in form of an adjacency list
 * @param adj_prev  the corresponding inverse adjacency list
 * @param s         the source node of G
 * @param t         the sink node of G
 * @param indegree  vector of size |V| containing the input degree for each node
 *
 * @return          tuple<level_vector, edge_flow, node_flow> with
 *                  level_vector: a vector containing the level of each node
 * 				edge_flow, node_flow: a flow network that represents the path cover
 */
tuple<level_vector,edge_flow,node_flow> calc_init_path_cover_level_sh(adj_list &adj, adj_list &adj_prev, size_t s,size_t t, vector<size_t> & indegree) {
	size_t n = adj.size();
	size_t l = 0;
	size_t p = 0;
	deque<size_t> Ql;
	Ql.push_back(s);
	
	level_vector level(n);
	level[s] = 0;
	
	cand_set R(n);
	path_ends path(0);
    
	edge_flow adj_f(n);
	node_flow v_f(n);
	
	unsigned int maxsize = 0;

	while (!Ql.empty()) {	
		//prepare Q for next level
		deque<size_t> Ql1(0);	   
		//prepare buckets
		const unsigned int bucket_count = maxsize + 5;
		vector<deque<size_t>> buckets (bucket_count);	
		
		for (size_t v : Ql) {  		
			// calculate R_v
			for (size_t u : adj_prev[v]) {
				for(size_t i : R[u]){
					size_t last = path[i];
					if(level[last] < level[u] || last == u )
						R[v].emplace(i);
				}
			}
            
            
			//sort in buckets
			if (maxsize < R[v].size()) 
				maxsize = R[v].size();
			
			if(R[v].size() < bucket_count)
				buckets[R[v].size()].push_back(v);
			else
				buckets.back().push_back(v);
		}
          
          //traverse buckets
		for(deque<size_t> B : buckets) { 
			for(size_t v : B){
				//choose path to cover i
				for(auto it=R[v].begin(); it != R[v].end(); it=R[v].begin()){
					if(level[path[*it]] == l)
						R[v].erase(it);
					else
						break;
				}            
            
				
				if (R[v].empty()) {  
					// add new path id
					R[v].insert(p++);
					path.push_back(v);
					fill_flow_s_v(s,v,adj_prev, adj_f, v_f); 
				}else{
					size_t i = *(R[v].begin());
					// find path from last to v
					size_t last = path[i];			
					fill_flow_u_v(last,v,adj_prev, level ,adj_f, v_f, i, R);
					path[i] = v;
				}

				for (size_t w : adj[v]) {
					indegree[w]--;
					if (!indegree[w]) {
						level[w] = l + 1;
						Ql1.push_back(w);
					}
				}
			}
			
		}
		l++;
		swap(Ql,Ql1);
	}
	
	for(unsigned int i = 0; i < path.size() ; i++)
		fill_flow_s_v(path[i],t,adj_prev, level ,adj_f, v_f, i, R);    
	
	return make_tuple(level,adj_f,v_f);
}


/**
 * Calculate the level for each node and a path cover that will be represented n a flow network.
 * 
 * initial path cover found with greedy set cover (Kousmanen et al)
 *
 * @param adj       DAG G in form of an adjacency list
 * @param adj_prev  the corresponding inverse adjacency list
 * @param s         the source node of G
 * @param t         the sink node of G
 * @param indegree  vector of size |V| containing the input degree for each node
 *
 * @return          tuple<level_vector, edge_flow, node_flow> with
 *                  level_vector: a vector containing the level of each node
 * 				edge_flow, node_flow: a flow network that represents the path cover
 */
tuple<level_vector,edge_flow,node_flow> calc_init_path_cover_gsc(adj_list &adj, adj_list &adj_prev, size_t s, size_t t, vector<size_t> &indegree){
	size_t n = adj.size();
	size_t l = 0;
	deque<size_t> Ql;
	Ql.push_back(s);
	level_vector level(n);
	vector<size_t> top_order(n);
	unsigned int top_order_next = 0;
	level[s] = 0;

	//calculate ml and a top_order
	while (!Ql.empty()) {		
	    	//prepare Q for next level
		deque<size_t> Ql1(0);
		for(size_t v : Ql){
			top_order[top_order_next++] = v;
			
			for (size_t w : adj[v]) {
				indegree[w]--;
				if (!indegree[w]) {
					level[w] = l + 1;
					Ql1.push_back(w);
				}
			}
		}
		l++;
		swap(Ql,Ql1);
    }

	edge_flow e_f(n);
	node_flow v_f(n);
    
	 //vector m and u of kousmanen paper
	vector<size_t> m(n,1);
	vector<size_t> u(n,1);
	
	while(true){ //uncovered nodes
		vector<size_t> suc(n);
		//calculate vector u
		for(auto v = top_order.rbegin(); v != top_order.rend(); v++){//traverse nedes in reverse top order
			int max_u_w = 0;
			for (size_t w : adj[*v]){
				if(u[w] > max_u_w || max_u_w == 0){
					max_u_w = u[w]; //max
					suc[*v] = w; //arg_max
				}				
			}
			u[*v] = m[*v] + max_u_w; //update u[v]
		}
		
		if(u[s] == 0) break; // no nodes uncovered
		
		//tracing back path
		size_t v = s;
		while(v != t){
			m[v] = 0; //v is used now
			v_f[v]++; //flow goes over this node
			size_t w = suc[v];			
			e_f[v][w]++; //flow goes over edge (w,v)
			v = w;
		}
		m[t] = 0;		
	}
    return make_tuple(level,e_f,v_f);
}


void fill_flow_s_v(size_t s,size_t v, adj_list &adj_prev, edge_flow &e_f, node_flow &v_f){
	v_f[v]++;
	while(v != s){
		size_t x = adj_prev[v].front();
		e_f[x][v]++;
		v_f[x]++;
		v = x;
	}
}

void fill_flow_u_v(size_t u,size_t v, adj_list &adj_prev, level_vector &level ,edge_flow &e_f, node_flow &v_f, size_t p, cand_set &R){
	while(v != u){
		v_f[v]++; //Flow over v
		for( size_t x : adj_prev[v] ){
			if(x==u || (level[x] > level[u] && R[x].count(p) ) ){
				e_f[x][v]++;
				v = x;
				break;
			}
		}
	}
}


/**
 * 3. minimize_pathCover:
 * Minimize the flow capacities by searching paths from t to s.
 *
 * @param adj       adjacency list of DAG G
 * @param adj_prev  inverse adjacency list of DAG G
 * @param s         the source node of G
 * @param t         the sink node of G
 * @param adj_f     flow capacities for each edge
 * @param vf        flow capacities for each node
 *
 * @return          tuple<adj_cap, node_cap> with
 *                  adj_cap:  minimized flow capacities for each edge
 *                  node_cap: minimized capacities for each node
 */
tuple<edge_flow, node_flow> minimize_flow(adj_list &adj, adj_list &adj_prev, size_t s, size_t t, edge_flow &e_f, node_flow &v_f, size_t capacity) {
	deque<pair<size_t, bool> > Q;  // Queue with pairs of {node, side}
	map<pair<size_t,bool>, pair<size_t,bool> > succ;
	Q.push_front({t,in});
	succ[{t,in}] = {t,out};
	while (!Q.empty()) { //find a path in residual network
		size_t v;
		bool v_side;
		tie(v, v_side) = Q.front();
		Q.pop_front();
     
		if (v == s && v_side == out) {// path completed -> remove flow on path
			v_f[v]--;
			while (v != t) {
				size_t w;
				bool w_side;
				tie(w, w_side) = succ[ {v,v_side}]; // remove one flow from (v,w)
				if (v == w) {  // same node but other side
					if (w_side == out) // v_in to v_out
						v_f[w]--;
					else  // w_side == in --> v_out to v_in
						v_f[w]++;                    
				} else {  // between different nodes
					if (w_side == in)  // v_out to w_in
						e_f[v][w]--;
					else  // w_side == out --> v_in to w_out
						e_f[w][v]++;
					
				}
				v = w;
				v_side = w_side;
			}
			succ.clear();
			deque<pair<size_t, bool> >().swap(Q);
			Q.push_front({t,in});
			succ[ {t,in}] = {t,out};
			continue;  // search for more paths
		}
		
		if (v_side == in) {
			for (size_t u : adj_prev[v]) {
			// use edge u_out -> v_in
				if (e_f[u][v] > 0 && !succ.count({u,out})) {
					Q.push_front({u,out});
					succ[ {u,out}] = {v,v_side};
				}
			}
			// use edge v_in -> v_out
			if (v_f[v] < capacity && !succ.count({v,out})) {
				Q.push_front({v,out});
				succ[ {v,out}] = {v,v_side};
			}
		} else {  // v_side = out
			// use edge v_out -> v_in
			if (v_f[v] > 1 && !succ.count({v,in})) {
				Q.push_front({v,in});
				succ[ {v,in}] = {v,v_side};
			}
			for (size_t w : adj[v]) {
				// use edge v_out -> w_in
				if (e_f[v][w] < capacity && !succ.count({w,in})) {
					Q.push_front({w,in});
					succ[ {w,in}] = {v,v_side};
				}
			}
		}
	}
	return make_tuple(e_f, v_f);
}


/**
 * Extract the paths of the flow network.
 * Calculate the pid coordinate for each node.
 *
 * @param adj       adjacency list of DAG G
 * @param s         the source node of G
 * @param t         the sink node of G
 * @param adj_f     minimized flow capacities for each edge
 *
 * @return          tuple<p, pid> with
 *                  p:   minimal number of needed paths for the path cover
 *                  pid: pid coordinate for each node
 */
vector<size_t> get_paths(adj_list &adj, size_t s, size_t t, size_t width, edge_flow &e_f) {
	size_t n = adj.size();
	vector<size_t> pid(n,-1);
	pid[0] = 0;
	pid[n-1] = 0;

	size_t p = 0;
	
	
	//iterate over paths
	for(size_t p = 0; p < width; p++){
		stringstream pathcover;
		pathcover << p+IDX_ONE << ": ";
		//search a path
		size_t v = s;
		while(v != t) {
			
			size_t w = s; // w = next node in path
			
			for (size_t z : adj[v]) 
				if (e_f[v][z] > 0 && pid[z] == (size_t) -1){ //prefer uncovered nodes
					w = z;
					break;
				}
				
			if(w == s)
				for (size_t z : adj[v]) 
					if (e_f[v][z] > 0){
						w = z;
						break;
					}
			
			e_f[v][w]--;					
			if(pid[v] == (size_t) -1)
				pid[v] = p;
				
			pathcover << v+IDX_ONE << ", ";
			v = w;				
		}
		pathcover << t+IDX_ONE << endl;
		PATHCOVER_STREAM << pathcover.str();
	}
	
	return pid;
}



/**
 * Extracts the DAG G from the GFA-FILE.
 * Stores additional information for each node.
 *
 * @param filename  path to the GFA-FILE
 * @param source 	name of the source node
 * @param sink 		name of the sink node
 *
 * @return          tuple<vertices, edge_list, node_name, name_idx> with
 *                  vertices:   number of vertices in DAG G
 *                  edge_list:  DAG G in form of an edge list
 *                  node_name:  base name of each node
 *                  name_idx:   seq_id of each node
					source_id: 	id of the source node
					sink_id: 	id of the sink node
 */
tuple<size_t, edge_list, vector<char>, vector<string>, size_t, size_t> read_gfa(string filename, string source, string sink) {
    vector<pair<size_t, size_t> > edges;
    map<string,pair<size_t,size_t> > seq_startEnd;
    vector<char> node_name;
    vector<string> node_idx;
    size_t s = source == "" ? NOT_SET : NOT_FOUND;
    size_t t = sink == "" ? NOT_SET : NOT_FOUND;

    GFAKluge gg = GFAKluge();
    gg.parse_gfa_file(filename);
    auto seq = gg.get_name_to_seq();
    auto seq_edges = gg.get_seq_to_edges();

    // add nodes
    size_t offset = 0;
    for (auto elem : seq) {
        string index = elem.first;
        size_t len = elem.second.length;
        size_t start = offset;
        size_t end = start + len - 1;
        seq_startEnd[index] = {start,end};

        if (index == source) {
            s = start;
        }
        if (index == sink) {
            t = end;
        }

        string name = elem.second.sequence;
        for (size_t i = 0; i < name.size(); i++) {
            node_name.emplace_back(name[i]);
            node_idx.emplace_back(index);
        }
        for (size_t i = 1; i < len; i++) {
            edges.emplace_back(offset + i - 1, offset + i);
        }
        offset += len;
    }

    // add edges
    for (auto map_entry : gg.get_seq_to_edges()) {
        auto seq_edges = map_entry.second;
        for (auto seq_edge : seq_edges) {
            string seq_from = seq_edge.source_name;
            string seq_to = seq_edge.sink_name;
            size_t from = seq_startEnd[seq_from].second;
            size_t to = seq_startEnd[seq_to].first;
            edges.emplace_back(from,to);
        }
    }
    return make_tuple(offset, edges, node_name, node_idx, s, t);
}



/**
 * User interface by command line.
 * Current available options:
 *   --pc=<filepath/stdout>
 *   --coords=<filepath/stdout>
 *   --dot=<filepath/stdout>
 *   --one
 *   --usecoords
 *   --source=<id>
 *   --sink=<id>
 *
 * @param argc      number of programm parameters
 * @param argv      list of programm parameters
 */
int main(int argc, char **argv) {
	filebuf pathcover_file_buffer;
	filebuf coords_file_buffer;
	filebuf dot_file_buffer;
	filebuf log_file_buffer;
	ostream pathcover_file(NULL);
	ostream coords_file(NULL);
	ostream dot_file(NULL);
	ostream log_file(NULL);
	string gfa_filename = "";
	unsigned int init_method = INIT_LEVEL;
	bool silent = false;
	string source = "";
	string sink = "";
	time_t time0 = time(NULL);

	for (int i = 1; i < argc; i++) {
		string input = argv[i];
		if (input[0] == '-' || input[1] == '-') {  // parse options
			input.erase(input.begin(), find_if(input.begin(), input.end(), [](int ch) {return ch != '-';}));
			size_t eq_pos = input.find("=");

			if (input == "help" || input == "h") {
				size_t sep = 32;
				cout << "Usage: " << argv[0] << " [OPTION]... GFA-FILE" << endl
					<< "Parses the graph stated in the GFA-FILE (gfa-format neccesary)" << endl
					<< "and calculates the maxLevel-coordinate for every base. " << endl
					<< endl
					<< "Options: " << endl
					// path cover file: --pc=<filepath/stdout>
					<< setw(sep) << left << "  --pc=<filepath/stdout>"
					<< "write the found path cover to the stated file " << endl
					<< setw(sep) << " " << "or to stdout" << endl

					// coords file: --coords=<filepath/stdout>
					<< setw(sep) << left << "  --coords=<filepath/stdout>"
					<< "write the calculated maxLevel coordinates to " << endl
					<< setw(sep) << " " << "the stated file or to stdout" << endl

					// dot file: --dot=<filepath/stdout>
					<< setw(sep) << left << "  --dot=<filepath/stdout>"
					<< "write the graph (dot-format) including the " << endl
					<< setw(sep) << " " << "calculated maxLevel coordinates to the " << endl
					<< setw(sep) << " " << "stated file or to stdout" << endl

					// dot file: --log=<filepath>
					<< setw(sep) << left << "  --log=<filepath/stdout>"
					<< "write log to the stated file " << endl
					<< setw(sep) << " " << "or to stdout" << endl

					// one indexed option
					<< setw(sep) << left << "  --one"
					<< "use one indexed numbers" << endl

					// without improved algorithm
					<< setw(sep) << left << "  --sh"
					<< "use algorithm with 'simple heuristic'=buckets sort of R lists before choosing the path id" << endl

					// Kmousmanen et al.
					<< setw(sep) << left << "  --gsc"
					<< "use algorithm with greedy set cover to find initial path cover" << endl

					<< setw(sep) << left << "  --silent"
					<< "do not print on command line" << endl

					// usecoords option
					<< setw(sep) << left << "  --usecoords"
					<< "use maxLevel coordinates for dot-graph" << endl
					<< setw(sep) << " " << "(need '-Kfdp' flag when using graphviz)" << endl

					// source option
					<< setw(sep) << left << "  --source=<id>"
					<< "set source node id (default value = 0)" << endl

					// sink option
					<< setw(sep) << left << "  --sink=<id>"
					<< "set sink node id (default value = n)" << endl

					<< "If in more than one argument stdout is stated, the information will be printed" << endl
					<< "consecutive without an empty line in between in the following order: " << endl
					<< "  1. path cover information" << endl
					<< "  2. maxLevel coordinates information" << endl
					<< "  3. dot graph information" << endl
					<< endl
					<< "Example:" << endl
					<< argv[0] << " --pc=pathCover.txt --coords=coords.txt --dot=graph.dot graph.gfa" << endl
					<< endl
					<< "Hint: dot-files can be visualized with graphviz using the following command:" << endl
					<< "  dot DOT-FILE -Tpdf > PDF-FILE" << endl
					;
				return 0;
			} else if (input == "one") {  // one option
				IDX_ONE = 1;
				continue;
			} else if (input == "usecoords") {  // usecoords option
                USE_COORD = true;
                continue;
			
			} else if (input == "gsc") {  // usecoords option
				init_method = INIT_GSC;
				continue;
			} else if (input == "sh") {  // usecoords option
				init_method = INIT_LEVEL_PLUS_SH;
				continue;
			} else if (input == "silent") {  // usecoords option
				silent = true;
				continue;
			} else if (eq_pos != string::npos) {
				string typ = input.substr(0, eq_pos);
				string arg = input.substr(eq_pos+1, input.size());
				if (typ == "pc") {  // path cover - file
					if (arg == "stdout") {
						pathcover_file.rdbuf(cout.rdbuf());
					} else {
						pathcover_file_buffer.open(arg, ios_base::out);
						pathcover_file.rdbuf(&pathcover_file_buffer);
					}
				} else if (typ == "coords") {  // coords - file
					if (arg == "stdout") {
					coords_file.rdbuf(cout.rdbuf());
					} else {
						coords_file_buffer.open(arg, ios_base::out);
						coords_file.rdbuf(&coords_file_buffer);
					}
				} else if (typ == "dot") {  // dot - file
					if (arg == "stdout") {
						dot_file.rdbuf(cout.rdbuf());
					} else {
						dot_file_buffer.open(arg, ios_base::out);
						dot_file.rdbuf(&dot_file_buffer);
					}
				} else if (typ == "log") {  // log filepath
					if (arg == "stdout") {
						log_file.rdbuf(cout.rdbuf());
					} else {
						log_file_buffer.open(arg, ios_base::out | ios_base::app);
						log_file.rdbuf(&log_file_buffer);
					}
				} else if (typ == "source") {  // source option
					source = arg;
				} else if (typ == "sink") {  // sink option
					sink = arg;
				} else {
					// unknown option
					cout << argv[0] << ": Unknown option '" << input << "'" << endl;
					cout << "Try '" << argv[0] << " --help' for more information." << endl;
					exit(EXIT_FAILURE);
				}
			} else {  // unknown option
				cout << argv[0] << ": Unknown option '" << input << "'" << endl;
				cout << "Try '" << argv[0] << " --help' for more information." << endl;
				exit(EXIT_FAILURE);
			}
		} else {  // read path for GFA-FILE
			if (gfa_filename != "") {  // multiple GFA-FILEs given
				cout << argv[0] << ": More than one GFA-FILE stated." << endl;
				cout << "Try '" << argv[0] << " --help' for more information." << endl;
				exit(EXIT_FAILURE);
			} else {
				gfa_filename = input;
			}
		}
	}
	if (gfa_filename == "") {  // check whether GFA-FILE is given
		cout << argv[0] << ": No GFA-FILE stated." << endl;
		cout << "Try '" << argv[0] << " --help' for more information." << endl;
		exit(EXIT_FAILURE);
	}
	struct stat info;
	if (stat(gfa_filename.c_str(), &info) != 0) {  // GFA-FILE doesn't exist
		cout << argv[0] << ": GFA-FILE doesn't exist." << endl;
		exit(EXIT_FAILURE);
	} else if (info.st_mode & S_IFDIR) {  // GFA-FILE is a directory
		cout << argv[0] << ": GFA-FILE is a directory." << endl;
		exit(EXIT_FAILURE);
	}

	// parse GFA-FILE and run <pg_coords>-algorithm
	size_t n, s, t;
	edge_list edges;
	
	if(!silent) cout << "read gfa..." << flush;
	time_t time1 = time(NULL);
	tie(n, edges, NODE_NAME, NODE_IDX, s, t) = read_gfa(gfa_filename, source, sink);
	if (s == NOT_FOUND) {
		cout << "Couldn't find the source node \"" << source << "\"" << endl;
		exit(EXIT_FAILURE);
	} else if (t == NOT_FOUND) {
		cout << "Couldn't find the sink node \"" << sink << "\"" << endl;
		exit(EXIT_FAILURE);
	}
	if (s == NOT_SET) s = 0; // default value for s = 0
	if (t == NOT_SET) t = n - 1; // default value for t = n

	
	time_t time2 = time(NULL);
	
	if(!silent) cout << "\rTime for reading gfa: " << difftime(time2,time1) << " s" << endl;
	LOG_STREAM << gfa_filename << '\t';
	
	
	
	calculate_coordinate_system(n, edges, s, t, init_method, silent);
	
	// write information to stated files or to stdout
	pathcover_file << PATHCOVER_STREAM.str();
	pathcover_file.flush();
	pathcover_file_buffer.close();
	coords_file << COORDS_STREAM.str();
	coords_file.flush();
	coords_file_buffer.close();
	dot_file << DOT_STREAM.str();
	dot_file.flush();
	dot_file_buffer.close();
	
	time1 = time(NULL);
	if(!silent) cout << "\rTime overall: " << difftime(time1,time0) << " s" << endl;
	LOG_STREAM << difftime(time1,time0) << endl;
	log_file << LOG_STREAM.str();
	log_file.flush();
	log_file_buffer.close();	
	exit(EXIT_SUCCESS);
}
